/* This software is licensed by the MIT License, see LICENSE file */
/*                                Copyright © 2022 Gregory Lirent */

#include "../../include/extra/dict.h"
#include "../__internal/assert.h"
#include "../__internal/rbtree.h"

#ifndef LIBCDSB_SRC_DICT_INCLUDE_H
#define LIBCDSB_SRC_DICT_INCLUDE_H

#define CAPACITY_BLOCK   100
#define REBUILD_POINT_MAX 0.65
#define REBUILD_POINT_MIN  1

typedef struct libcdsb_dict_node {
    struct libcdsb_dict_node* left;
    struct libcdsb_dict_node* right;
    struct libcdsb_dict_node* parent;

    vnode_t key;
    short   colored;
    u8_t    key_type;
    u8_t    val_type;
    vnode_t value;
} dnode_t;

static_assert(offsetof(struct libcdsb_rbtree_node, left) == offsetof(struct libcdsb_dict_node, left), "Implementation assert");
static_assert(offsetof(struct libcdsb_rbtree_node, right) == offsetof(struct libcdsb_dict_node, right), "Implementation assert");
static_assert(offsetof(struct libcdsb_rbtree_node, parent) == offsetof(struct libcdsb_dict_node, parent), "Implementation assert");
static_assert(offsetof(struct libcdsb_rbtree_node, colored) == offsetof(struct libcdsb_dict_node, colored), "Implementation assert");
static_assert(offsetof(struct libcdsb_rbtree_node, value) == offsetof(struct libcdsb_dict_node, key), "Implementation assert");

#define dnode_empty           ((dnode_t*)LIBCDSB_RBTREE_NODE_EMPTY)
#define dnode_create(k, p, c) ((dnode_t*)libcdsb_rbtree_node_create(k, (rbnode_t*)p, c, sizeof(dnode_t)))
#define dnode_fixup(r, n)     libcdsb_rbtree_node_fixup((rbnode_t**)(r), (rbnode_t*)(n))
#define dnode_delete(r, n)    (dnode_t*)libcdsb_rbtree_node_delete((rbnode_t**)(r), (rbnode_t*)(n))

#define dnode_is_empty(n) ((n) == dnode_empty)
#define dnode_is_root(n)  dnode_is_empty((n)->parent)

extern void libcdcb_dnode_free(dnode_t* x, void* not_used);

#define dnode_free libcdcb_dnode_free

#endif /* LIBCDSB_SRC_MAP_INCLUDE_H */
